#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2011 Centrify Corporation.
#
# Machine mapper script that checks if FIPS mode is supported.
#
#  This script checks domain functional level. If domain functional level
#  is not Windows Server 2008 or above, then stop adclient.
#
#
# Parameters: <map|unmap> mode
#   map|unmap   action to take
#   mode        mode (not used)
#
# Exit value:
#   0   Normal
#   1   Error
#   2   usage
#
##############################################################################

use strict;

use lib '/usr/share/centrifydc/perl';

use CentrifyDC::Config;
use CentrifyDC::GP::Args;
use CentrifyDC::GP::General qw(:debug RunCommand);

my $DO_STOP_ADCLIENT = '/var/centrifydc/reg/do_stop_adclient';

my $FIPS_PARAM = 'fips.mode.enable';

sub ScheduleStopAdclient();
sub IsFipsSupported();
sub IsFipsAllowed();
sub ApplyFipsSetting();



# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

sub ScheduleStopAdclient()
{
    INFO_OUT("Schedule stop of adclient");

    open(FH, ">$DO_STOP_ADCLIENT");
    close(FH);
    chmod(0600, $DO_STOP_ADCLIENT);
}

#
# check if current OS supports FIPS. Only Mac and RedHat Linux supports FIPS.
#
#   return: 1       - FIPS supported
#           0       - FIPS not supported
#           undef   - failed
#
sub IsFipsSupported()
{
    my $supported = 0;

    my $system = `uname -s`;
    chomp $system;

    if ($system eq 'Darwin')
    {
        $supported = 1;
    }
    elsif ($system eq 'Linux')
    {
        if (-x '/usr/bin/lsb_release')
        {
            # first try to identify the system using lsb
            my $release = `/usr/bin/lsb_release -i`;
            chomp $release;

            if ($release =~ m/RedHat/)
            {
                $supported = 1;
            }
        }
    }

    return $supported;
}

#
# check domain functional level to see if FIPS mode is allowed. FIPS is allowed
# only when domainFunctionality is no less than 3 (i.e. Windows Server 2008).
#
#   return: 1       - FIPS allowed
#           0       - FIPS not allowed
#           undef   - failed
#
sub IsFipsAllowed()
{
    my $is_fips_allowed = 0;

    my $cmd = "/usr/share/centrifydc/bin/ldapsearch -m -H ldap:// -b '' -s base -LLL domainFunctionality";

    my ($ret, $output) = RunCommand($cmd);
    if (! defined($ret) or $ret ne 0)
    {
        ERROR_OUT("ldapsearch failed");
        return undef;
    }

    # remove all the junks and get attribute
    if ($output =~ m/^domainFunctionality:/m)
    {
        $output =~ m/^domainFunctionality:\s*(.*)$/m;
        $output = $1;
    }
    else
    {
        $output = undef;
    }

    if (! defined($output))
    {
        ERROR_OUT("Cannot get domain functional level");
        return undef;
    }

    DEBUG_OUT("Domain functional level: %d", $output);

    if ($output < 3)
    {
        DEBUG_OUT("FIPS mode not allowed at current domain functional level %d.", $output);
        $is_fips_allowed = 0;
    }
    else
    {
        DEBUG_OUT("FIPS mode allowed at current domain functional level %d.", $output);
        $is_fips_allowed = 1;
    }

    return $is_fips_allowed;
}

#
# stop adclient if FIPS mode is enabled in centrifydc.conf but current domain
# functional level doesn't support FIPS.
#
# first get FIPS setting from centrifydc.conf, then check if FIPS mode is
# allowed in current domain functional level. If FIPS mode is enabled but
# is not allowed in current domain functional level, then stop adclient.
#
#   return: 1       - successful
#           2       - no need to do anything
#           undef   - failed
#
sub ApplyFipsSetting()
{
    my $setting = $CentrifyDC::Config::properties{$FIPS_PARAM};

    # if FIPS mode is not enabled, do nothing
    ($setting eq 'true') or return 2;

    # check domain functional level to make sure FIPS mode is allowed.
    my $fips_allowed = IsFipsAllowed();
    
    if (! defined($fips_allowed))
    {
        ERROR_OUT("Cannot determine if FIPS mode is allowed or not.");
        return undef;
    }
    elsif ($fips_allowed)
    {
        return 2;
    }
    else
    {
        INFO_OUT("FIPS mode is not allowed in current domain functional level. Stop adclient.");
        ScheduleStopAdclient();
    }

    return 1;
}

#
# map
#
#   return: 1       - successful
#           undef   - failed
#
sub Map()
{
    return ApplyFipsSetting();
}



# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $args = CentrifyDC::GP::Args->new('machine');

if (! IsFipsSupported())
{
    DEBUG_OUT("FIPS mode is not supported on current platform.");
    exit(0);
}

my $ret = 0;

$args->isMap() ? ($ret = Map()) : ($ret = 1);

$ret or FATAL_OUT();


