#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2007-2008 Centrify Corporation.
#
# Configure sshd settings.
#
#  This mapper script will update sshd_config, then verify it using "sshd -t".
#  If new configuration is valid, it will restart sshd; else original settings
#  will be restored.
#
#  If multiple openssh package are installed on one system, there is no easy
#  way to figure out which one is actually in use. This mapper will search for
#  Centrify openssh first, if not found, it will try to search other
#  pre-defined location. It will only use the first file it found.
#
#  Notice: This mapper will try to find a safe way to restart sshd (i.e. not
#  disconnect already established session), so it will not use killall -HUP.
#  If there's no safe way to restart sshd, it will simply create a log entry,
#  saying that "ssh policy will not take effect until ssh daemon is restarted".
#
##############################################################################

use strict;

use lib '/usr/share/centrifydc/perl';

use CentrifyDC::GP::Args;
use CentrifyDC::GP::General qw(:debug RunCommand);
use CentrifyDC::GP::Mapper;

my $is_sshd_config_valid = 0;



# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#
# Try to restart sshd. If sshd is not running or sshd configuration file
# bad option, then sshd won't be restarted.
#
# Return value is not used by caller.
#
#   ret:    0   - successful
#           1   - failed to restart sshd or configuration file error
#           2   - cannot find convenient way to restart sshd
#
sub ::restart_sshd()
{
    if ($is_sshd_config_valid == 0)
    {
        WARN_OUT("sshd configuration file contains bad option. Rollback to original setting. Will not restart sshd.");
        return 1;
    }

    my $system = `uname -s`;
    chomp $system;
    my $command;
    my $sshd_start_script;
    my $ret = 0;

    my @sshd_start_scripts = ("/etc/init.d/centrify-sshd",  # Centrify ssh
                        "/sbin/init.d/centrify-sshd", # Centrify ssh for SunOS/HP-UX
                        "/etc/init.d/sshd",           # SunOS/IRIX/IRIX64/Linux
                        "/sbin/init.d/secsh");        # HP-UX
    foreach my $sshd_start_script (@sshd_start_scripts)
    {
        if (-x $sshd_start_script)
        {
            # Instead of restart, we stop sshd first and then start.
            # If sshd is not running, it will not be started.
            $command = "$sshd_start_script stop && $sshd_start_script start";
            last;
        }
    }

    if (! defined($command))
    {
        if ($system eq "AIX")
        {
            if (-x "/usr/bin/stopsrc" && -x "/usr/bin/startsrc")
            {
                my $subsystem_name = "sshd";
                if (-x "/usr/share/centrifydc/sbin/sshd")
                {
                    # if Centrify openssh is installed, use it.
                    $subsystem_name = "centrify-sshd";
                }
                $command = "/usr/bin/stopsrc -s $subsystem_name && /usr/bin/startsrc -s $subsystem_name";
            }
        }
        elsif ($system eq "IRIX" || $system eq "IRIX64")
        {
            my $pidfile = "/etc/ssh/sshd.pid";
            if (-e $pidfile && -x "/sbin/cat")
            {
                $command = "kill -HUP `/sbin/cat $pidfile`";
            }
            else
            {
                # killall will disconnect established session, so we'll
                # not use it.
                # $command = "killall -HUP sshd";
                $command = "echo ssh daemon configuration file is updated, new settings will take effect once sshd restarts.";
            }
        }
        elsif ($system eq "SunOS")
        {
            my $pidfile = "/var/run/sshd.pid";
            if (-x "/usr/sbin/svcadm")
            {
                $command = "/usr/sbin/svcadm restart ssh";
            }
            elsif (-e $pidfile && -x "/usr/bin/cat")
            {
                $command = "kill -HUP `/usr/bin/cat $pidfile`";
            }
            else
            {
                # pkill will disconnect established session.
                # $command = "pkill -HUP sshd";
                $command = "echo ssh daemon configuration file is updated, new settings will take effect once ssh daemon restarts.";
            }
        }
        elsif ($system eq "Darwin")
        {
            $command = "echo No need to restart ssh daemon on Mac OS X.";
        }
    }

    if (defined($command))
    {
        DEBUG_OUT("Restart ssh daemon to apply ssh policy: [$command]");
        my $rc = RunCommand($command);
        if (! defined($rc) or $rc ne '0')
        {
            DEBUG_OUT("ssh daemon restart failed. ssh policy will not take effect until ssh daemon is restarted.");
            $ret = 1;
        }
        else
        {
            DEBUG_OUT("ssh daemon restarted successfully. ssh group policy is applied.");
            $ret = 0;
        }
    }
    else
    {
        WARN_OUT("Cannot find a convenient method to restart ssh daemon. ssh policy will not take effect until ssh daemon is restarted.");
        $ret = 2;
    }

    return $ret;
}

#
# Try to find sshd and validate the modified sshd_config using "sshd -t".
# If file or keys are invalid, will revert sshd_config.
#
#   ret:    0   - correct
#           1   - incorrect
#           2   - cannot find sshd to verify
#
sub ::verify_sshd_config()
{
    my $ret = 0;
    my $system = `uname -s`;
    chomp $system;
    my $command;

    my @sshd_paths = ("/usr/share/centrifydc/sbin/sshd", # Centrify ssh
                        "/usr/sbin/sshd",           # AIX/HP-UX/SunOS/Darwin/Linux
                        "/usr/lib/ssh/sshd",        # SunOS
                        "/usr/local/sbin/sshd",     # AIX/SunOS
                        "/usr/freeware/sbin/sshd",  # IRIX/IRIX64
                        "/opt/ssh/sbin/sshd");      # HP-UX
    foreach my $sshd_path (@sshd_paths)
    {
        if (-x $sshd_path)
        {
            $command = $sshd_path . " -t";
            last;
        }
    }

    if (defined($command))
    {
        DEBUG_OUT("Validate sshd configuration file and keys: $command");
        my $rc = RunCommand($command);
        if (! defined($rc) or $rc ne '0')
        {
            WARN_OUT("sshd configuration file contains bad option. ssh policy will not be applied.");
            $ret = 1;
        }
        else
        {
            $is_sshd_config_valid = 1;
            DEBUG_OUT("sshd configuration file is correct.");
            $ret = 0;
        }
    }
    else
    {
        WARN_OUT("Cannot find ssh daemon to verify configuration file. ssh policy will not be applied.");
        $ret = 2;
    }

    return $ret;
}



# >>> DATA >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $file = {
    'hierarchy_separator' => '.',
    'list_expr' => ', *| +',
    'list_separator' => ', ',
    'match_expr' => [
      '/([^\s:=]+)[:=]?\s*(.*)/',
    ],
    'comment_markers' => [ '#', ],
    'comment_match_expr' => [
      '/#([^\s:=]+)[:=]?\s*(.*)/',
    ],

    'path' => [
      '/etc/centrifydc/ssh/sshd_config',    # Centrify ssh
      '/usr/local/etc/sshd_config',         # AIX/SunOS
      '/etc/ssh/sshd_config',               # AIX/SunOS/IRIX/IRIX64/Linux
      '/etc/sshd_config',                   # Darwin/SunOS
      '/opt/ssh/etc/sshd_config',           # HP-UX
    ],
    'lock' => '/tmp/gp.sshd_config.lock',

    'verify_command' => '::verify_sshd_config()',

    'post_command' => '::restart_sshd()',

    'value_map' => {

      'UsePAM' => {
        'reg_class' => 'machine',

        'reg_key' => 'Software/Policies/Centrify/ssh',
        'reg_value' => 'UsePAM',
        'reg_type'   => ['REG_SZ'],
        'value_type' => 'named',
      },

      'Banner'   => {
        'active'     => '$value_map->{Banner_hidden}{reg_data} ne "0"',
        'default_data' => '/etc/issue',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'Banner',
        'reg_type'   => ['REG_SZ'],
        'value_type' => 'named',
      },

      'Banner_hidden'   => {
        'active'     => '0',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'ssh.banner.enabled',
        'reg_type'   => ['REG_DWORD'],
        'value_type' => 'named',
      },

      'ClientAliveCountMax'   => {
        'default_data' => '3',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'ClientAliveCountMax',
        'reg_type'   => ['REG_DWORD'],
        'value_type' => 'named',
      },

      'ClientAliveInterval'   => {
        'default_data' => '0',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'ClientAliveInterval',
        'reg_type'   => ['REG_DWORD'],
        'value_type' => 'named',
      },

      'AllowGroups'   => {
        'active'     => '$value_map->{AllowGroups_hidden}{reg_data} ne "0"',
        'default_data' => '*',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'AllowGroups',
        'reg_type'   => ['REG_SZ'],
        'value_type' => 'named',
      },

      'AllowGroups_hidden'   => {
        'active'     => '0',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'ssh.allowgroups.enabled',
        'reg_type'   => ['REG_DWORD'],
        'value_type' => 'named',
      },

      'AllowUsers'   => {
        'active'     => '$value_map->{AllowUsers_hidden}{reg_data} ne "0"',
        'default_data' => '*',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'AllowUsers',
        'reg_type'   => ['REG_SZ'],
        'value_type' => 'named',
      },

      'AllowUsers_hidden'   => {
        'active'     => '0',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'ssh.allowusers.enabled',
        'reg_type'   => ['REG_DWORD'],
        'value_type' => 'named',
      },

      'DenyGroups'   => {
        'active'     => '$value_map->{DenyGroups_hidden}{reg_data} ne "0"',
        'default_data' => '*',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'DenyGroups',
        'reg_type'   => ['REG_SZ'],
        'value_type' => 'named',
      },

      'DenyGroups_hidden'   => {
        'active'     => '0',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'ssh.denygroups.enabled',
        'reg_type'   => ['REG_DWORD'],
        'value_type' => 'named',
      },

      'DenyUsers'   => {
        'active'     => '$value_map->{DenyUsers_hidden}{reg_data} ne "0"',
        'default_data' => '*',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'DenyUsers',
        'reg_type'   => ['REG_SZ'],
        'value_type' => 'named',
      },

      'DenyUsers_hidden'   => {
        'active'     => '0',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'ssh.denyusers.enabled',
        'reg_type'   => ['REG_DWORD'],
        'value_type' => 'named',
      },

      'GSSAPIAuthentication' => {
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'GSSAPIAuthentication',
        'reg_type'   => ['REG_SZ'],
        'value_type' => 'named',
      },

      'GSSAPIKeyExchange' => {
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'GSSAPIKeyExchange',
        'reg_type'   => ['REG_SZ'],
        'value_type' => 'named',
      },

      'LoginGraceTime'   => {
        'default_data' => '120',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'LoginGraceTime',
        'reg_type'   => ['REG_DWORD'],
        'value_type' => 'named',
      },

      'LogLevel'   => {
        'default_data' => 'INFO',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'LogLevel',
        'reg_type'   => ['REG_SZ'],
        'value_type' => 'named',
      },

      'PermitRootLogin'   => {
        'default_data' => 'yes',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'PermitRootLogin',
        'reg_type'   => ['REG_SZ'],
        'value_type' => 'named',
      },

      'Protocol' => {
        'default_data' => '1,2',
        'reg_class'  => 'machine',
        'reg_key'    => 'Software/Policies/Centrify/ssh',
        'reg_value'  => 'Protocol',
        'reg_type'   => ['REG_SZ'],
        'value_type' => 'named',
      },


    },

    'write_data' => '$value $data\n',
};



# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $args = CentrifyDC::GP::Args->new();

#
# need to undef library path. old sshd (4.1) uses libcrypto.a. if we don't
# undef library path, sshd will use /usr/share/centrifydc/lib/libcrypto.a
# instead of the correct one, and will fail to load.
#
exists($ENV{LIBPATH}) && undef $ENV{LIBPATH};
exists($ENV{LD_LIBRARY_PATH}) && undef $ENV{LD_LIBRARY_PATH};

if ($args->isMap())
{
    CentrifyDC::GP::Mapper::Map($file, $args->user());
}
else
{
    CentrifyDC::GP::Mapper::UnMap($file, $args->user());
}

